#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the lange project.
#  Please respect the license - more about this in the section (*) below.
#
#  lange is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  lange is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with lange.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

import decimal as dec
import math

from lange.tricks import detect_precision


# TODO implement (lazy) reverse: ap[. . .].r
# TODO implement lazy direct access to item/slice (use math formulas to know value without iterating)


class Progression:
    _l = None

    def __init__(self, op, null, prod_f, div_f, pow_f, bins_f, args, maxdigits=28):
        self.__str__ = self.__repr__
        self.args = args
        self.maxdigits = maxdigits
        self.prod_f, self.div_f, self.pow_f = prod_f, div_f, pow_f
        self.op = op
        self.decctx = dec.Context()
        null = self.decctx.create_decimal(null)
        if Ellipsis in args:
            if not (3 <= len(args) <= 4) or args[2] is not Ellipsis:
                raise Exception("Among 3 or 4 arguments, '...' should be the third.")

            # Detect level of precision and enforce it for the rest of this object life.
            end = 0 if len(args) == 3 else args[-1]
            self.precision = max(
                detect_precision(args[0], maxdigits),
                detect_precision(args[1], maxdigits),
                detect_precision(end, maxdigits),
            )
            self.decctx.prec = self.precision

            # Protect all provided numbers from floating point inequality issues (e.g. 0.8 - 0.6 != 0.2).
            self.start = self.decctx.create_decimal(args[0])
            self.second = self.decctx.create_decimal(args[1])
            self.end = self.decctx.create_decimal(math.inf if len(args) == 3 else args[-1])

            # Derived attributes.
            self.step = div_f(self.second, self.start)
            self.cast = int if all(isinstance(v, int) for v in args[:2]) and int(self.step) == self.step else float
            bins = bins_f(self.start, self.end, self.step).to_integral_exact(rounding=dec.ROUND_FLOOR)
            self.n = bins + self.decctx.create_decimal(1)

            def g():
                s, i = self.start, 0
                while i < self.n:
                    yield self.cast(s)
                    s = prod_f(s, self.step)
                    i += 1

            self.gen = g
        elif len(args) == 0:
            self.start, self.step, self.end = None, None, None
            self.n = 0
            self.gen = lambda: iter(())
            self.cast = lambda x: x
        elif 1 <= len(args) <= 2:
            self.start = self.decctx.create_decimal(args[0])
            self.step = null if len(args) == 1 else self.div_f(self.decctx.create_decimal(args[1]), self.start)
            self.end = self.decctx.create_decimal(args[-1])
            self.n = len(args)
            self.gen = lambda: iter(args)
            self.cast = int if all(isinstance(v, int) for v in args) and int(self.step) == self.step else float
        else:
            self.start, self.step, self.end = None, None, None
            self.n = len(args)
            self.gen = lambda: iter(args)
            self.cast = int if all(isinstance(v, int) for v in args) else float

    @property
    def l(self):
        """Return progression evaluated as a list.

        Usage:
            >>> from lange import ap
            >>> len(ap[1, 2, ..., 5])
            5
            >>> ap[1, 2, ..., 5].l
            [1, 2, 3, 4, 5]
        """
        if self._l is None:
            self._l = list(self)
        return self._l

    def __iter__(self):
        return self.gen()

    def __getitem__(self, item):
        if isinstance(item, slice):
            if self.start is None:
                return self.__class__(*list(self)[item])
            item_start = item.start or self.decctx.create_decimal(0)
            item_step = item.step or self.decctx.create_decimal(1)
            start = self.decctx.create_decimal(self[item_start])
            step = self.div_f(self.decctx.create_decimal(self[item_start + item_step]), start)
            second = self.prod_f(start, step)
            end = self.prod_f(start, self.pow_f(step, (item.stop - item_start - 1) // item_step))
            args = start, second, ..., end
            return self.__class__(*args)  # REMINDER: calling child class.

        if self.start is None:
            return self.__class__(list(self)[item]) if self.n > 0 else self.__class__()

        if not (0 <= item < self.n):
            raise Exception(f"Index {item} outside valid range [0; {self.n}[.")
        return self.cast(self.prod_f(self.start, self.pow_f(self.step, item)))

    def __len__(self):
        return int(self.n)

    def __repr__(self):
        if self.n < 4 or None in [self.end, self.start, self.step]:
            return f"[{' '.join(map(str, self))}]"
        try:
            end = self.cast(self.end)
        except OverflowError:
            end = "∞"
        return f"[{self.cast(self.start)} {self.cast(self.prod_f(self.start, self.step))} .{self.op}. {end}]"

    def __invert__(self):
        return list(self)



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the lange project.
#  Please respect the license - more about this in the section (*) below.
#
#  lange is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  lange is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with lange.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

from lange.abs.progression import Progression


class GP(Progression):
    """Geometric Progression. A flexible inductive sequence of numbers based on Haskell interval notation.

    maxdigits: maximum amount of digits verified during precision detection
                (includes numbers to the left and to the right of the decimal separator, if any)

    Usage:
        >>> print(GP(5))
        [5]
        >>> print(GP(5, 1))
        [5 1]
        >>> gp = GP(0.3, 0.6, ..., 6)
        >>> gp
        [0.3 0.6 .*. 6.0]
        >>> list(gp)
        [0.3, 0.6, 1.2, 2.4, 4.8]
        >>> gp[1]
        0.6
        >>> list(GP(0.2, 4.1, 0.9))
        [0.2, 4.1, 0.9]
        >>> gp = GP(0.3, 0.6, ...)
        >>> gp
        [0.3 0.6 .*. inf]
        >>> GP(0.3, 0.6, ...)[:8]
        [0.3 0.6 .*. 38.4]
        >>> list(GP(0.3, 0.6, ...)[:8])
        [0.3, 0.6, 1.2, 2.4, 4.8, 9.6, 19.2, 38.4]
        >>> GP()
        []
        >>> list(GP())
        []

    Usage (square brackets syntax):
        >>> from lange import gp
        >>> gp[0.3, 0.6, ..., 2]
        [0.3 0.6 1.2]
        >>> gp[1, 2]
        [1 2]
        >>> list(gp[1, 2])
        [1, 2]
        >>> gp[1]
        [1]
        >>> list(gp[1])
        [1]
        >>> gp[3, 1, 6, 7, 0]
        [3 1 6 7 0]
        >>> list(gp[3, 1, 6, 7, 0])
        [3, 1, 6, 7, 0]
        >>> gp[4, 5, 2, 9, 4, 7, 3, 3][2:6]
        [2 9 4 7]
        >>> gp[4, 5, 2, 9, 4, 7, 3, 3][5]
        [7]
        >>> list(gp[3, 2, 6666, 7, 8, 6, 5555, 7, 3, 9][2:9:4])
        [6666, 5555]
        >>> print(gp[0.1, 0.2, ...][2:8])
        [0.4 0.8 .*. 12.8]
        >>> gp[0.1, 0.2, ...][2:8:2]
        [0.4 1.6 6.4]
        >>> list(gp[2,3,...,14])
        [2.0, 3.0, 4.5, 6.75, 10.125]
        >>> gp[1,2,...]
        [1 2 .*. ∞]
        >>> gp[1,1.00000000001,...][1_000_000_000_000]
        22026.465793705393
    """

    def __init__(self, *args, maxdigits=28):
        super().__init__(
            "*",
            1,
            lambda a, b: a * b,
            lambda a, b: a / b,
            lambda a, b: a**b,
            lambda a, b, c: (b / a).log10() / c.log10(),
            args,
            maxdigits,
        )

    def __reduce__(self):  # pragma: no cover
        return GP, self.args + (self.maxdigits,)



# Based on:
# https://stackoverflow.com/questions/3018758/determine-precision-and-ret-of-particular-number-in-python


#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the lange project.
#  Please respect the license - more about this in the section (*) below.
#
#  lange is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  lange is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with lange.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

import decimal as dec
from functools import partial
from math import nan


def detect_precision(x, maxdigits=28):
    """
    Detect the precision (include digits before and after the decimal separator).

    The checking only works if 'max_digits' is large enough.

    Usage:
        >>> detect_precision(3)
        1
        >>> detect_precision(3.7)
        2
        >>> detect_precision(3.0)
        1
        >>> detect_precision(30.0e-4)
        4
        >>> detect_precision(30.0e4)
        6

    Based on:
    https://stackoverflow.com/questions/3018758/determine-precision-and-ret-of-particular-number-in-python
    """
    if x is ...:
        return 0
    ctx = dec.Context()
    ctx.prec = maxdigits
    d1 = ctx.create_decimal(repr(float(x)))
    txt = format(d1, "f")
    l = len(txt) - 1
    if int(x) == x:
        l -= 1
    return l


def list2progression(lst, maxdigits=28):
    """Convert list representing A. or G. progression to lange

    >>> list2progression([0,1,2,...,9])
    [0 1 .+. 9]
    >>> list2progression([1,2,4,...,16])
    [1 2 .*. 16]

    Parameters
    ----------
    lst

    Returns
    -------

    """
    if len(lst) < 3:  # pragma: no cover
        raise Exception(
            f"Cannot guess if you want an arithmetic or a geometric projection. Provide 3 numbers, not {len(lst)}."
        )

    # Protect diffs and ratios from floating point inequality issues (e.g. 0.8 - 0.6 != 0.2).
    precision = max(map(partial(detect_precision, maxdigits=maxdigits), lst))
    decctx = dec.Context()
    decctx.prec = precision
    lst_dec = [decctx.create_decimal(x) for x in lst if x is not ...]

    # Calculate diffs and ratios.
    try:
        diff1 = lst_dec[1] - lst_dec[0]
        diff2 = lst_dec[2] - lst_dec[1]
        ratio1 = nan if lst_dec[0] == 0 else lst_dec[1] / lst_dec[0]
        ratio2 = nan if lst_dec[1] == 0 else lst_dec[2] / lst_dec[1]
    except:  # pragma: no cover
        raise InconsistentLange(f"Cannot identify whether this is a G. or A. progression: {lst}")
    newlst = lst[0:2] + lst[3:]

    if diff1 == diff2:
        from lange.ap import AP

        return AP(*newlst)
    elif ratio1 == ratio2:
        from lange.gp import GP

        return GP(*newlst)
    else:  # pragma: no cover
        raise InconsistentLange(
            f"Cannot identify whether this is a G. or A. Progression: {lst}", diff1, diff2, ratio1, ratio2
        )


class InconsistentLange(Exception):
    pass



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the lange project.
#  Please respect the license - more about this in the section (*) below.
#
#  lange is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  lange is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with lange.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

from lange.ap import AP
from lange.gp import GP
from lange.tricks import list2progression


class APwithBrackets:
    def __getitem__(self, item):
        """Helper class to allow overriding square brackets.

        Import object 'ap' from module 'lange' to use it implicitly.

        Usage:
            >>> from lange import ap
            >>> ap[0.6, 0.8, ..., 2]
            [0.6 0.8 .+. 2.0]
            >>> list(ap[0.6, 0.8, ..., 2])
            [0.6, 0.8, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0]
        """
        if isinstance(item, tuple):
            return AP(*item)
        else:
            return AP(item)


class GPwithBrackets:
    """Helper class to allow overriding square brackets.

    Import object 'gp' from module 'lange' to use it implicitly.

    Usage:
        >>> from lange import gp
        >>> gp[0.3, 0.6, ..., 2]
        [0.3 0.6 1.2]
        >>> list(gp[0.3, 0.6, ..., 2])
        [0.3, 0.6, 1.2]
    """

    def __getitem__(self, item):
        if isinstance(item, tuple):
            return GP(*item)
        else:
            return GP(item)


class AnyPwithBrackets:
    """Helper class to allow overriding square brackets.

    Import object 'pr' from module 'lange' to use it implicitly.

    Usage:
        >>> from lange import pr
        >>> pr[0.6, 0.8, 1, ..., 2]
        [0.6 0.8 .+. 2.0]
        >>> list(pr[0.6, 0.8, 1, ..., 2])
        [0.6, 0.8, 1.0, 1.2, 1.4, 1.6, 1.8, 2.0]
        >>> pr[0.3, 0.6, 1.2, ..., 2]
        [0.3 0.6 1.2]
        >>> list(pr[0.3, 0.6, 1.2, ..., 3])
        [0.3, 0.6, 1.2, 2.4]
    """

    def __getitem__(self, item):
        return list2progression(item)
